@@ -1,50 +1,41 @@
+from __future__ import annotations
+
 import asyncio
-from typing import Any, Callable, cast, Generator, Optional
+from ssl import SSLError
+from typing import Any, Generator, Optional
 
-from .context import Context
 from .task_group import TaskGroup
+from .worker_context import WorkerContext
 from ..config import Config
 from ..events import Closed, Event, RawData, Updated
 from ..protocol import ProtocolWrapper
-from ..typing import ASGIFramework
+from ..typing import AppWrapper
 from ..utils import parse_socket_addr
 
-MAX_RECV = 2 ** 16
-
-
-class EventWrapper:
-    def __init__(self) -> None:
-        self._event = asyncio.Event()
-
-    async def clear(self) -> None:
-        self._event.clear()
-
-    async def wait(self) -> None:
-        await self._event.wait()
-
-    async def set(self) -> None:
-        self._event.set()
+MAX_RECV = 2**16
 
 
 class TCPServer:
     def __init__(
         self,
-        app: ASGIFramework,
+        app: AppWrapper,
         loop: asyncio.AbstractEventLoop,
         config: Config,
+        context: WorkerContext,
         reader: asyncio.StreamReader,
         writer: asyncio.StreamWriter,
     ) -> None:
         self.app = app
         self.config = config
+        self.context = context
         self.loop = loop
         self.protocol: ProtocolWrapper
         self.reader = reader
         self.writer = writer
         self.send_lock = asyncio.Lock()
-        self.timeout_lock = asyncio.Lock()
+        self.idle_lock = asyncio.Lock()
 
-        self._keep_alive_timeout_handle: Optional[asyncio.Task] = None
+        self._idle_handle: Optional[asyncio.Task] = None
 
     def __await__(self) -> Generator[Any, None, None]:
         return self.run().__await__()
@@ -63,11 +54,11 @@ class TCPServer:
                 alpn_protocol = "http/1.1"
 
             async with TaskGroup(self.loop) as task_group:
-                context = Context(task_group)
                 self.protocol = ProtocolWrapper(
                     self.app,
                     self.config,
-                    cast(Any, context),
+                    self.context,
+                    task_group,
                     ssl,
                     client,
                     server,
@@ -75,7 +66,7 @@ class TCPServer:
                     alpn_protocol,
                 )
                 await self.protocol.initiate()
-                await self._update_keep_alive_timeout()
+                await self._start_idle()
                 await self._read_data()
         except OSError:
             pass
@@ -88,35 +79,32 @@ class TCPServer:
                 try:
                     self.writer.write(event.data)
                     await self.writer.drain()
-                except (BrokenPipeError, ConnectionResetError):
+                except (ConnectionError, RuntimeError):
                     await self.protocol.handle(Closed())
         elif isinstance(event, Closed):
             await self._close()
             await self.protocol.handle(Closed())
         elif isinstance(event, Updated):
-            pass  # Triggers the keep alive timeout update
-        await self._update_keep_alive_timeout()
+            if event.idle:
+                await self._start_idle()
+            else:
+                await self._stop_idle()
 
     async def _read_data(self) -> None:
-        while True:
+        while not self.reader.at_eof():
             try:
-                data = await self.reader.read(MAX_RECV)
+                data = await asyncio.wait_for(self.reader.read(MAX_RECV), self.config.read_timeout)
             except (
-                BrokenPipeError,
-                ConnectionRefusedError,
-                ConnectionResetError,
+                ConnectionError,
                 OSError,
                 asyncio.TimeoutError,
                 TimeoutError,
+                SSLError,
             ):
                 await self.protocol.handle(Closed())
                 break
             else:
-                if data == b"":
-                    await self._update_keep_alive_timeout()
-                    break
                 await self.protocol.handle(RawData(data))
-                await self._update_keep_alive_timeout()
 
     async def _close(self) -> None:
         try:
@@ -127,24 +115,33 @@ class TCPServer:
         try:
             self.writer.close()
             await self.writer.wait_closed()
-        except (BrokenPipeError, ConnectionResetError):
+        except (BrokenPipeError, ConnectionResetError, RuntimeError):
             pass  # Already closed
 
-    async def _update_keep_alive_timeout(self) -> None:
-        async with self.timeout_lock:
-            if self._keep_alive_timeout_handle is not None:
-                self._keep_alive_timeout_handle.cancel()
-            self._keep_alive_timeout_handle = None
-            if self.protocol.idle:
-                self._keep_alive_timeout_handle = self.loop.create_task(
-                    _call_later(self.config.keep_alive_timeout, self._timeout)
-                )
+        await self._stop_idle()
 
-    async def _timeout(self) -> None:
+    async def _initiate_server_close(self) -> None:
         await self.protocol.handle(Closed())
         self.writer.close()
 
+    async def _start_idle(self) -> None:
+        async with self.idle_lock:
+            if self._idle_handle is None:
+                self._idle_handle = self.loop.create_task(self._run_idle())
 
-async def _call_later(timeout: float, callback: Callable) -> None:
-    await asyncio.sleep(timeout)
-    await asyncio.shield(callback())
+    async def _stop_idle(self) -> None:
+        async with self.idle_lock:
+            if self._idle_handle is not None:
+                self._idle_handle.cancel()
+                try:
+                    await self._idle_handle
+                except asyncio.CancelledError:
+                    pass
+            self._idle_handle = None
+
+    async def _run_idle(self) -> None:
+        try:
+            await asyncio.wait_for(self.context.terminated.wait(), self.config.keep_alive_timeout)
+        except asyncio.TimeoutError:
+            pass
+        await asyncio.shield(self._initiate_server_close())
