@@ -1,14 +1,29 @@
+from __future__ import annotations
+
 from enum import auto, Enum
 from time import time
 from typing import Awaitable, Callable, Optional, Tuple
 from urllib.parse import unquote
 
-from .events import Body, EndBody, Event, Request, Response, StreamClosed
+from .events import Body, EndBody, Event, InformationalResponse, Request, Response, StreamClosed
 from ..config import Config
-from ..typing import ASGIFramework, Context
-from ..utils import build_and_validate_headers, suppress_body, UnexpectedMessage, valid_server_name
+from ..typing import (
+    AppWrapper,
+    ASGISendEvent,
+    HTTPResponseStartEvent,
+    HTTPScope,
+    TaskGroup,
+    WorkerContext,
+)
+from ..utils import (
+    build_and_validate_headers,
+    suppress_body,
+    UnexpectedMessageError,
+    valid_server_name,
+)
 
 PUSH_VERSIONS = {"2", "3"}
+EARLY_HINTS_VERSIONS = {"2", "3"}
 
 
 class ASGIHTTPState(Enum):
@@ -23,9 +38,10 @@ class ASGIHTTPState(Enum):
 class HTTPStream:
     def __init__(
         self,
-        app: ASGIFramework,
+        app: AppWrapper,
         config: Config,
-        context: Context,
+        context: WorkerContext,
+        task_group: TaskGroup,
         ssl: bool,
         client: Optional[Tuple[str, int]],
         server: Optional[Tuple[str, int]],
@@ -37,14 +53,15 @@ class HTTPStream:
         self.closed = False
         self.config = config
         self.context = context
-        self.response: dict
-        self.scope: dict
+        self.response: HTTPResponseStartEvent
+        self.scope: HTTPScope
         self.send = send
         self.scheme = "https" if ssl else "http"
         self.server = server
         self.start_time: float
         self.state = ASGIHTTPState.REQUEST
         self.stream_id = stream_id
+        self.task_group = task_group
 
     @property
     def idle(self) -> bool:
@@ -59,7 +76,7 @@ class HTTPStream:
             self.scope = {
                 "type": "http",
                 "http_version": event.http_version,
-                "asgi": {"spec_version": "2.1"},
+                "asgi": {"spec_version": "2.1", "version": "3.0"},
                 "method": event.method,
                 "scheme": self.scheme,
                 "path": unquote(path.decode("ascii")),
@@ -69,12 +86,16 @@ class HTTPStream:
                 "headers": event.headers,
                 "client": self.client,
                 "server": self.server,
+                "extensions": {},
             }
             if event.http_version in PUSH_VERSIONS:
-                self.scope["extensions"] = {"http.response.push": {}}
+                self.scope["extensions"]["http.response.push"] = {}
+
+            if event.http_version in EARLY_HINTS_VERSIONS:
+                self.scope["extensions"]["http.response.early_hint"] = {}
 
             if valid_server_name(self.config, event):
-                self.app_put = await self.context.spawn_app(
+                self.app_put = await self.task_group.spawn_app(
                     self.app, self.config, self.scope, self.app_send
                 )
             else:
@@ -90,9 +111,9 @@ class HTTPStream:
         elif isinstance(event, StreamClosed):
             self.closed = True
             if self.app_put is not None:
-                await self.app_put({"type": "http.disconnect"})
+                await self.app_put({"type": "http.disconnect"})  # type: ignore
 
-    async def app_send(self, message: Optional[dict]) -> None:
+    async def app_send(self, message: Optional[ASGISendEvent]) -> None:
         if self.closed:
             # Allow app to finish after close
             return
@@ -125,6 +146,19 @@ class HTTPStream:
                         raw_path=message["path"].encode(),
                     )
                 )
+            elif (
+                message["type"] == "http.response.early_hint"
+                and self.scope["http_version"] in EARLY_HINTS_VERSIONS
+                and self.state == ASGIHTTPState.REQUEST
+            ):
+                headers = [(b"link", bytes(link).strip()) for link in message["links"]]
+                await self.send(
+                    InformationalResponse(
+                        stream_id=self.stream_id,
+                        headers=headers,
+                        status_code=103,
+                    )
+                )
             elif message["type"] == "http.response.body" and self.state in {
                 ASGIHTTPState.REQUEST,
                 ASGIHTTPState.RESPONSE,
@@ -157,7 +191,7 @@ class HTTPStream:
                         await self.send(EndBody(stream_id=self.stream_id))
                         await self.send(StreamClosed(stream_id=self.stream_id))
             else:
-                raise UnexpectedMessage(self.state, message["type"])
+                raise UnexpectedMessageError(self.state, message["type"])
 
     async def _send_error_response(self, status_code: int) -> None:
         await self.send(
