@@ -90,6 +90,11 @@ class HTTPServerPushEvent(TypedDict):
     headers: Iterable[Tuple[bytes, bytes]]
 
 
+class HTTPEarlyHintEvent(TypedDict):
+    type: Literal["http.response.early_hint"]
+    links: Iterable[bytes]
+
+
 class HTTPDisconnectEvent(TypedDict):
     type: Literal["http.disconnect"]
 
@@ -180,6 +185,7 @@ ASGISendEvent = Union[
     HTTPResponseStartEvent,
     HTTPResponseBodyEvent,
     HTTPServerPushEvent,
+    HTTPEarlyHintEvent,
     HTTPDisconnectEvent,
     WebsocketAcceptEvent,
     WebsocketSendEvent,
@@ -196,19 +202,7 @@ ASGISendEvent = Union[
 ASGIReceiveCallable = Callable[[], Awaitable[ASGIReceiveEvent]]
 ASGISendCallable = Callable[[ASGISendEvent], Awaitable[None]]
 
-
-class ASGI2Protocol(Protocol):
-    # Should replace with a Protocol when PEP 544 is accepted.
-
-    def __init__(self, scope: Scope) -> None:
-        ...
-
-    async def __call__(self, receive: ASGIReceiveCallable, send: ASGISendCallable) -> None:
-        ...
-
-
-ASGI2Framework = Type[ASGI2Protocol]
-ASGI3Framework = Callable[
+ASGIFramework = Callable[
     [
         Scope,
         ASGIReceiveCallable,
@@ -216,7 +210,8 @@ ASGI3Framework = Callable[
     ],
     Awaitable[None],
 ]
-ASGIFramework = Union[ASGI2Framework, ASGI3Framework]
+WSGIFramework = Callable[[dict, Callable], Iterable[bytes]]
+Framework = Union[ASGIFramework, WSGIFramework]
 
 
 class H2SyncStream(Protocol):
@@ -282,10 +277,13 @@ class Event(Protocol):
     async def wait(self) -> None:
         ...
 
+    def is_set(self) -> bool:
+        ...
+
 
 class WorkerContext(Protocol):
     event_class: Type[Event]
-    terminated: bool
+    terminated: Event
 
     @staticmethod
     async def sleep(wait: Union[float, int]) -> None:
@@ -299,7 +297,7 @@ class WorkerContext(Protocol):
 class TaskGroup(Protocol):
     async def spawn_app(
         self,
-        app: ASGIFramework,
+        app: AppWrapper,
         config: Config,
         scope: Scope,
         send: Callable[[Optional[ASGISendEvent]], Awaitable[None]],
@@ -319,3 +317,15 @@ class TaskGroup(Protocol):
 class ResponseSummary(TypedDict):
     status: int
     headers: Iterable[Tuple[bytes, bytes]]
+
+
+class AppWrapper(Protocol):
+    async def __call__(
+        self,
+        scope: Scope,
+        receive: ASGIReceiveCallable,
+        send: ASGISendCallable,
+        sync_spawn: Callable,
+        call_soon: Callable,
+    ) -> None:
+        ...
